@@ -3,6 +3,7 @@
 
 import json
 from inorbit_edge import __version__ as inorbit_edge_version
+from inorbit_edge.types import Pose, SpatialTolerance
 import os
 import logging
 import paho.mqtt.client as mqtt
@@ -10,6 +11,8 @@ from urllib.parse import urlsplit
 import socks
 import ssl
 import threading
+import yaml
+
 from inorbit_edge.inorbit_pb2 import (
     CustomDataMessage,
     KeyValueCustomElement,
@@ -22,9 +25,17 @@ from inorbit_edge.inorbit_pb2 import (
     Echo,
     CustomScriptCommandMessage,
     CustomScriptStatusMessage,
+    CustomCommandRosMessage,
     CameraMessage,
 )
 from inorbit_edge.video import CameraStreamer, Camera
+from inorbit_edge.missions import MissionsModule
+from inorbit_edge.commands import (
+    COMMAND_INITIAL_POSE,
+    COMMAND_NAV_GOAL,
+    COMMAND_CUSTOM_COMMAND,
+    COMMAND_MESSAGE,
+)
 import time
 import requests
 import math
@@ -51,13 +62,10 @@ MQTT_NAV_GOAL_GOAL = "ros/loc/nav_goal"
 MQTT_NAV_GOAL_MULTI = "ros/loc/goal_path"
 MQTT_INITIAL_POSE = "ros/loc/set_pose"
 MQTT_CUSTOM_COMMAND = "custom_command/script/command"
+MQTT_CUSTOM_COMMAND_MESSAGE = "custom_command/ros"
 MQTT_SCRIPT_OUTPUT_TOPIC = "custom_command/script/status"
 MQTT_IN_CMD = "in_cmd"
 
-# InOrbit commands
-COMMAND_INITIAL_POSE = "initialPose"
-COMMAND_NAV_GOAL = "navGoal"
-COMMAND_CUSTOM_COMMAND = "customCommand"
 # InOrbit modules
 INORBIT_MODULE_CAMERAS = "RosImageAgentlet"
 # CustomCommand execution status
@@ -68,26 +76,31 @@ ROBOT_PATH_POINTS_LIMIT = 1000
 
 
 class RobotSession:
-    def __init__(self, robot_id, robot_name, api_key, **kwargs) -> None:
+    def __init__(self, robot_id, robot_name, api_key=None, **kwargs) -> None:
         """Initialize a robot session.
 
         Args:
             robot_id (str): ID of the robot.
-            robot_name (str): Robot name.
             api_key (str): API key for authenticating against InOrbit Cloud services.
+            robot_name (str): Robot name.
+            robot_key(str): Robot key for authenticating against InOrbit Cloud services
+            when using InOrbit Connect (https://connect.inorbit.ai/).
             endpoint (str): InOrbit URL. Defaults: INORBIT_CLOUD_SDK_ROBOT_CONFIG_URL.
             use_ssl (bool): Configures MQTT client to use SSL. Defaults: True.
         """
 
+        self.api_key = api_key
+
         self.logger = logging.getLogger(__class__.__name__)
 
         self.robot_id = robot_id
-        self.robot_name = robot_name
-        self.api_key = api_key
+        self.robot_key = kwargs.get("robot_key")
+        self.robot_name = kwargs.get("robot_name", robot_name)
         # The agent version is generated based on the InOrbit Edge SDK version
         self.agent_version = "{}.edgesdk_py".format(inorbit_edge_version)
         self.endpoint = kwargs.get("endpoint", INORBIT_CLOUD_SDK_ROBOT_CONFIG_URL)
-
+        # Track robot's current pose
+        self._last_pose = None
         # Use SSL by default
         self.use_ssl = kwargs.get("use_ssl", True)
 
@@ -144,12 +157,14 @@ class RobotSession:
         self.message_handlers = {}
 
         self.command_callbacks = []
+        self.missions_module = MissionsModule(self)
         self.camera_streamers = {}
         self.camera_streaming_on = False
         self.camera_streaming_mutex = threading.Lock()
 
         self.message_handlers[MQTT_INITIAL_POSE] = self._handle_initial_pose
         self.message_handlers[MQTT_CUSTOM_COMMAND] = self._handle_custom_command
+        self.message_handlers[MQTT_CUSTOM_COMMAND_MESSAGE] = self._handle_custom_message
         self.message_handlers[MQTT_NAV_GOAL_GOAL] = self._handle_nav_goal
         self.message_handlers[MQTT_IN_CMD] = self._handle_in_cmd
 
@@ -244,12 +259,16 @@ class RobotSession:
         self.logger.info("Fetching config for robot {}".format(self.robot_id))
         # get params from self
         params = {
-            "appKey": self.api_key,
             "robotId": self.robot_id,
             "hostname": self.robot_name,
             "agentVersion": self.agent_version,
         }
 
+        if self.robot_key:
+            params["robotKey"] = self.robot_key
+        elif self.api_key:
+            params["appKey"] = self.api_key
+
         # post request to fetch robot config
         response = requests.post(self.endpoint, data=params)
         response.raise_for_status()
@@ -290,6 +309,9 @@ class RobotSession:
         self.client.subscribe(
             topic=self._get_robot_subtopic(subtopic=MQTT_CUSTOM_COMMAND)
         )
+        self.client.subscribe(
+            topic=self._get_robot_subtopic(subtopic=MQTT_CUSTOM_COMMAND_MESSAGE)
+        )
         self.client.subscribe(
             topic=self._get_robot_subtopic(subtopic=MQTT_NAV_GOAL_GOAL)
         )
@@ -361,7 +383,7 @@ class RobotSession:
         theta = args[4]
 
         # Hand over to callback for processing, using the proper format
-        self._dispatch_command(
+        self.dispatch_command(
             command_name=COMMAND_INITIAL_POSE,
             args=[{"x": x, "y": y, "theta": theta}],
             execution_id=seq,  # NOTE: Using seq as the execution ID
@@ -373,12 +395,23 @@ class RobotSession:
         custom_script_msg = CustomScriptCommandMessage()
         custom_script_msg.ParseFromString(msg)
         # Hand over to callback for processing, using the proper format
-        self._dispatch_command(
+        self.dispatch_command(
             command_name=COMMAND_CUSTOM_COMMAND,
             args=[custom_script_msg.file_name, custom_script_msg.arg_options],
             execution_id=custom_script_msg.execution_id,
         )
 
+    def _handle_custom_message(self, msg):
+        """Handle incoming MQTT_CUSTOM_COMMAND_MESSAGE message."""
+
+        custom_command_message = CustomCommandRosMessage()
+        custom_command_message.ParseFromString(msg)
+        # Hand over to callback for processing, using the proper format
+        self.dispatch_command(
+            command_name=COMMAND_MESSAGE,
+            args=[custom_command_message.cmd],
+        )
+
     def _handle_nav_goal(self, msg):
         """Handle incoming MQTT_NAV_GOAL_GOAL message."""
 
@@ -389,7 +422,7 @@ class RobotSession:
         y = args[3]
         theta = args[4]
         # Hand over to callback for processing, using the proper format
-        self._dispatch_command(
+        self.dispatch_command(
             command_name=COMMAND_NAV_GOAL,
             args=[{"x": x, "y": y, "theta": theta}],
             execution_id=seq,  # NOTE: Using seq as the execution ID
@@ -439,12 +472,13 @@ class RobotSession:
         msg.image = image
         self.publish_protobuf(MQTT_SUBTOPIC_CAMERA_V2, msg)
 
-    def _dispatch_command(self, command_name, args, execution_id):
+    def dispatch_command(self, command_name, args, execution_id=None):
         """Executes registered command callbacks for a specific incoming command."""
         for callback in self.command_callbacks:
 
             def result_function(result_code):
-                return self.report_command_result(args, execution_id, result_code)
+                if execution_id is not None:
+                    return self.report_command_result(args, execution_id, result_code)
 
             # TODO: Implement progress reporting function
             def progress_function(output, error):
@@ -746,17 +780,34 @@ class RobotSession:
         msg.pos_y = y
         msg.yaw = yaw
         msg.frame_id = frame_id
+        self._last_pose = Pose(frame_id=frame_id, x=x, y=y, theta=yaw)
         self.publish_protobuf(MQTT_SUBTOPIC_POSE, msg)
 
-    def publish_key_values(self, key_values, custom_field="0"):
+    def reached_waypoint(self, waypoint: Pose, tolerance: SpatialTolerance):
+        if self._last_pose is None:
+            return False
+        return (
+            math.sqrt(
+                (self._last_pose.x - waypoint.x) ** 2
+                + (self._last_pose.y - waypoint.y) ** 2,
+            )
+            <= tolerance.positionMeters
+            and abs(self._last_pose.theta - waypoint.theta) % (2 * math.pi)
+            <= tolerance.angularRadians
+        )
+
+    def publish_key_values(self, key_values, custom_field="0", is_event=False):
         """Publish key value pairs
 
         Args:
             key_values (dict): Key value mappings to publish
             custom_field (str, optional): ID of the CustomData element. Defaults to "0".
+            is_event (bool): Events are not throttled
         """
 
-        if not self._should_publish_message(method="publish_key_values"):
+        if not is_event and not self._should_publish_message(
+            method="publish_key_values"
+        ):
             return None
 
         def convert_value(value):
@@ -937,12 +988,15 @@ class RobotSessionFactory:
         self.command_callbacks = []
         self.commands_paths_rules = []
 
-    def build(self, robot_id, robot_name):
+    def build(self, robot_id, robot_name="", **robot_config):
         """Builds a RobotSession object using the provided id and name.
         It also passes the robot_session_kw_args set when creating the factory to the
         RobotSession constructor.
         """
-        session = RobotSession(robot_id, robot_name, **self.robot_session_kw_args)
+
+        session = RobotSession(
+            robot_id, robot_name, **{**robot_config, **self.robot_session_kw_args}
+        )
 
         def build_callback(callback):
             def c(*args):
@@ -979,14 +1033,30 @@ class RobotSessionPool:
     """Pool of robot sessions that handles connections for many robots in an
     efficient way"""
 
-    def __init__(self, robot_session_factory):
+    def __init__(self, robot_session_factory, robot_config_yaml=None):
         """Creates the pool
         Args:
           - robot_session_factory: factory used to build individual RobotSession
           objects
         """
+
+        self.logger = logging.getLogger(__class__.__name__)
+
         self.robot_session_factory = robot_session_factory
         self.robot_sessions = {}
+        self.robot_config = {}
+
+        # If a robots config yaml file was provided, load it
+        if robot_config_yaml:
+            try:
+                with open(robot_config_yaml, "r") as config_yaml:
+                    self.robot_config = yaml.safe_load(config_yaml)
+            except Exception:
+                self.logger.error(
+                    f"Unable to load robots config yaml at {robot_config_yaml}"
+                )
+                raise
+
         self.getting_session_mutex = threading.Lock()
 
     def get_session(self, robot_id, robot_name=""):
@@ -999,8 +1069,14 @@ class RobotSessionPool:
         new_robot_session = not self.has_robot(robot_id)
         try:
             if new_robot_session:
+                # Get the config params for this robot_id
+                robot_config = self.robot_config.get(robot_id, {})
+                # If there is no robot name in the config yaml, use the one
+                # provided to this method.
+                if not robot_config.get("robot_name"):
+                    robot_config["robot_name"] = robot_name
                 self.robot_sessions[robot_id] = self.robot_session_factory.build(
-                    robot_id, robot_name
+                    robot_id, **robot_config
                 )
                 self.robot_sessions[robot_id].connect()
             return self.robot_sessions[robot_id]
